/*
  A client for the Strata server.
  Copyright (C) 2009, 2010 Red Hat, Inc

  Author(s): Gavin Romig-Koch <gavin@redhat.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
  This module has two primary sets of entry points: the first is the
  reportfile_ functions, the send_report_to_ functions.

  The reportfile_ functions are for building an 'report' in a file
  or block of memory.  The send_report_to functions are for http POST-ing
  a 'report' to a Strata server.

*/

// Throughout this code there are references to 'body'.  This is the
// _content_ of an HTTP request or response, and should probably be 
// renamed such throughout.

// This code could benifit from being rewritten in C++, and adding
// extern "C" bindings where useful.


#define _GNU_SOURCE 

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <locale.h>
#include <stdint.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>
#include <curl/curl.h>

#include "strata_client.h"

struct reportfile {
  char* tmpdir;
  xmlTextWriterPtr writer;
  xmlBufferPtr buf;
  int issignature;
};

static char* internal_error_string = NULL;

static char*
ssprintf( const char* format, ... ) {
  int r;
  va_list p;
  char* retval = NULL;

  va_start(p, format);
  r = vasprintf(&retval, format, p);
  va_end(p);

  if (r < 0) {
    internal_error_string = strdup("ssprintf: vasprintf failed");
    if (!internal_error_string) 
      internal_error_string = "ssprintf: vasprintf failed";
    return 0;
  }

  return retval;
}

static void
internal_error_printf( const char* format, ... ) {
  int r;
  va_list p;

  va_start(p, format);
  r = vasprintf(&internal_error_string, format, p);
  va_end(p);

  if (r < 0) {
    internal_error_string = strdup("internal_error_printf: vasprintf failed");
    if (!internal_error_string) 
      internal_error_string = "internal_error_printf: vasprintf failed";
  } 
}

static int 
acat_mem(const char** oldp, const void* add, int add_len) {

  if (!add_len) return 0;

  const char* old = *oldp;
  int old_len = old ? strlen(old) : 0;

  int newstr_len = old_len + add_len;
  char* newstr = (char*)malloc(newstr_len+1);

  if (!newstr) {
    internal_error_printf("malloc failed");
    return 1;
  }

  mempcpy(mempcpy(newstr,old,old_len),add,add_len);
  newstr[newstr_len] = 0;

  free((void*)old);
  *oldp = newstr;
  return 0;
}

static int 
acat_str(const char** oldp, const char* add) {

  int add_len = add ? strlen(add) : 0;
  if (!add_len) return 0;
  return acat_mem(oldp, add, add_len);
}

static int 
runv(const char *filename, const char *const argv[]) {
  int status;

  pid_t pid = fork ();
  if (pid == 0) {
    execv( filename, (char*const*)argv);
    _exit (EXIT_FAILURE);

  } else if (pid < 0) {
    status = -1;

  } else {
    if (waitpid (pid, &status, 0) != pid)
      status = -1;
  }

  if (status) {
    internal_error_printf("'%s' exited with status %d", filename, status);
    return -1;
  }

  return 0;
}

static int
copy_file(const char *src_name, const char *dst_name) {
  const char* const argv[] = { "cp", "--", src_name, dst_name, 0 };
  return runv("/bin/cp", argv);
}

static int 
remove_directory(const char* dir) {
  const char* const argv[] = { "rm", "-rf", "--", dir, 0 };
  return runv("/bin/rm", argv);
}

static char*
concat_path_file(const char *path, const char *filename) {
  // This adds concats the two names, adding or removing one slash
  // as necessary.  It does not handle the case that path ends
  // in, or filename starts with, multiple slashs.

  if (path == NULL && filename == NULL)
    return strdup("");

  else if (path == NULL)
    return strdup(filename);
 
  else if (filename == NULL)
    return strdup(path);

  int path_len = strlen(path);
  int filename_len = strlen(filename);

  if (path_len == 0 && filename_len == 0)
    return strdup("");

  else if (path_len == 0)
    return strdup(filename);
 
  else if (filename_len == 0)
    return strdup(path);

  const char* path_end = path + strlen(path) - 1;
  if ((*path_end == '/') != (*filename == '/'))
    // one has it, the other doesn't
    return ssprintf("%s%s", path, filename);
  else if (*filename == '/')
    // they both have it
    return ssprintf("%s%s", path, filename+1);
  else
    // neither has it
    return ssprintf("%s/%s", path, filename);
}


static int
write_rgn_to_file( const char *value, size_t size,
                   const char *path, const char *fname ) {

  char* ofile_name = concat_path_file(path, fname);
  if (!ofile_name)
    return -1;

  FILE *ofile = fopen(ofile_name, "w");
  if (!ofile) {
    internal_error_printf("Can't open '%s'", ofile_name);
    free((void*)ofile_name);
    return -1;
  }

  size_t value_size = (size == SIZE_MAX) ? strlen(value) : size;
  if (value_size != fwrite(value, 1, value_size, ofile)) {
    internal_error_printf("Can't write to '%s'", ofile_name);
    free((void*)ofile_name);
    return -1;
  }

  free((void*)ofile_name);
  fclose(ofile);
  return 0;
}

const char* 
strata_client_strerror( void ) {
  return internal_error_string;
}

//
// The purpose of all the reportfile_XXX functions is to 
// collect up the contents of a report or signature into a 
// single tarball or xml file.
// 


// 
// This allocates a reportfile_t structure and initializes it.
//
//   Set 'issignature' to true if you only want an xml file.
//   If 'issignature' is true, then any bindings which would 
//   be written to an external file are dropped.
//
reportfile_t* 
reportfile_start( int issignature ) {

  // create a new reportfile_t
  reportfile_t* file = (reportfile_t*)malloc(sizeof(reportfile_t));
  if (!file) {
    internal_error_printf("Error: reportfile_start: Could not allocate reportfile_t");
    return NULL;
  }

  file->issignature = issignature;

  // make a temp directory to work in
  file->tmpdir = strdup("/tmp/reportfileXXXXXX");
  if (!file->tmpdir) {
    reportfile_free(file);
    internal_error_printf("Error: reportfile_start: Could not allocate tmpdir");
    return NULL;
  }
  if (mkdtemp(file->tmpdir) == NULL) {
    reportfile_free(file);
    internal_error_printf("Can't mkdtemp a temporary directory in /tmp");
    return NULL;
  }

  { // create a 'content' dir within temp directory
    char* content_dir_name = concat_path_file(file->tmpdir, "content");
    if (!content_dir_name) return NULL;
    if (mkdir(content_dir_name, 0700)) {
      reportfile_free(file);
      free(content_dir_name);
      internal_error_printf("Can't mkdir '%s'\n", content_dir_name);
      return NULL;
    }
  }

  // set up a libxml 'buffer' and 'writer' to that buffer
  file->buf = xmlBufferCreate();
  if (file->buf == NULL) {
    reportfile_free(file);
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    return NULL;
  }
  file->writer = xmlNewTextWriterMemory(file->buf, 0);
  if (file->writer == NULL) {
    reportfile_free(file);
    internal_error_printf("strata_client: Error creating the xml writer\n");
    return NULL;
  }

  // start a new xml document
  int rc;
  rc = xmlTextWriterStartDocument(file->writer, NULL, NULL, NULL);
  if (rc < 0) {
    reportfile_free(file);
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    return NULL;
  }

  // start the document element 'report'
  rc = xmlTextWriterStartElement(file->writer, (const xmlChar*)"report");
  if (rc < 0) {
    reportfile_free(file);
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return NULL;
  }

  // add the 'xmlns' attribue
  rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"xmlns",
                                   (const xmlChar*)"http://www.redhat.com/gss/strata");
  if (rc < 0) {
    reportfile_free(file);
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return NULL;
  }

  return file;
}


static int
internal_reportfile_start_binding( reportfile_t* file, const char* name, 
                                   int isbinary, const char* filename ) {
  int rc;
  
  // open a binding element
  rc = xmlTextWriterStartElement(file->writer, (const xmlChar*)"binding");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return -1;
  }

  // add the 'name' attribue
  rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"name", (const xmlChar*)name);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return -1;
  }


  // add the 'filename' attribute if neede
  if (filename) {
    rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"fileName", (const xmlChar*)filename);
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }

  // add the 'type' attribute
  if (isbinary) {
    rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"type", (const xmlChar*)"binary");
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  else {
    rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"type", (const xmlChar*)"text");
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  return 0;
}

//
// Add a new binding to a report, for a binding whos value is represented
//   as either a null-terminated string or start+len region of memory.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   
//   'value' and 'size' hold the value to be bound to 'name'
//      If 'size' == SIZE_MAX then 'value' is treated as a null-terminated
//        string; otherwise 'value' the start of a region of memory 'size' 
//        bytes long.
//
//      Note that while it is valid to set 'isbinary' for a value that is
//        a null-terminated string, it probably is not correct.
//
static int
reportfile_add_binding_from_region( reportfile_t* file, 
                                    const char* name, 
                                    const void* value,
                                    size_t size,
                                    int isbinary, 
                                    const char* filename ) {
  int rc;

  // start the binding element with primary attributes
  rc = internal_reportfile_start_binding(file, name, isbinary, filename);
  if (rc < 0)
    return rc;

  // write the value of the binding, in the way that makes sense
  //   given it's value
  if (isbinary) {
    char* content_dir_name = concat_path_file(file->tmpdir, "content");
    if (!content_dir_name)
      return -1;
    rc = write_rgn_to_file(value, size, content_dir_name, name);
    free(content_dir_name);
    if (rc < 0)
      return rc;

    char* href_name = concat_path_file("content", name);
    if (!href_name)
      return -1;
    rc = xmlTextWriterWriteAttribute(file->writer, 
                                     (const xmlChar*)"href", 
                                     (const xmlChar*)href_name);
    free(href_name);
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  else {
    rc = xmlTextWriterWriteAttribute(file->writer, 
                                     (const xmlChar*)"value", 
                                     (const xmlChar*)value);
      if (rc < 0) {
        internal_error_printf
	  ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        return -1;
    }
  }

  rc = xmlTextWriterEndElement(file->writer);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterEndElement\n");
    return -1;
  }
  return 0;
}

// Add a new binding to a report, for a binding whos value is represented
//   as a null-terminated string.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   'value' is a null-terminated string to bound to name.
//      Note that while it is valid to set 'isbinary' for a value that is
//        a null-terminated string, it probably is not correct.
//
int
reportfile_add_binding_from_string( reportfile_t* file, 
                                    const char* name, 
                                    const char* value,
                                    int isbinary, 
                                    const char* filename ) {
  return reportfile_add_binding_from_region(file, name, 
                                            value, SIZE_MAX, 
                                            isbinary, filename);
}

// Add a new binding to a report, for a binding whos value is represented
//   as a null-terminated string.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   'value' is a null-terminated string to bound to name.
//      Note that while it is valid to set 'isbinary' for a value that is
//        a null-terminated string, it probably is not correct.
//
int
reportfile_add_binding_from_memory( reportfile_t* file, 
                                    const char* name, 
                                    const void* value,
                                    size_t size,
                                    int isbinary, 
                                    const char* filename ) {
  if (size == SIZE_MAX)
    {
      internal_error_printf(
          "reportfile_add_binding_from_memory: error 'size' too long");
      return -1;
    }
  return reportfile_add_binding_from_region(file, name, 
                                            value, size, 
                                            isbinary, filename);
}



//
// Add a new binding to a report, for a binding whos value is represented
//   as a named file.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   
//   'local_filename' is the name of the file containing the bindings value
//      'filename' and 'local_filename' can be the same, but don't need to
//         be the same.
//
int
reportfile_add_binding_from_namedfile( reportfile_t* file, const char* name, const char* local_filename, int isbinary, const char* filename ) {
  int rc;

  rc = internal_reportfile_start_binding(file, name, isbinary, filename);
  if (rc < 0)
    return rc;

  char* content_dir_name = concat_path_file(file->tmpdir, "content");
  if (!content_dir_name)
    return -1;
  char* ofile_name = concat_path_file(content_dir_name, name);
  free((void*)content_dir_name);
  if (!ofile_name)
    return -1;
  rc = copy_file(local_filename, ofile_name);
  free(ofile_name);
  if (rc < 0)
    return rc;

  
  char* href_name = concat_path_file("content", name);
  if (!href_name)
    return -1;
  rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"href", (const xmlChar*)href_name);

  free(href_name);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return -1;
  }

  rc = xmlTextWriterEndElement(file->writer);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterEndElement\n");
    return -1;
  }
  return 0;
}

//
// End the reportfile, and prepare it for delivery. 
// No more bindings can be added after this.
//
int
reportfile_end( reportfile_t* file ) {

  int rc = -1;

  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(file->writer);
  if (!rc) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    rc = -1;
  }

  xmlFreeTextWriter(file->writer);
  file->writer = NULL;
  return rc;
}

//
// Write the reportfile to 'outfile_name'
//
// Currently this only works if the reportfile is a Report.
//
int
reportfile_write_to_file( reportfile_t* file, const char* outfile_name ) {

  if (0) {
    internal_error_printf("error: reportfile_write_to_file called before reportfile_end");
    return -1;
  }

  if (file->issignature) {
    internal_error_printf("error: reportfile_write_to_file called on signature");
    return -1;
  }

  int rc;
  // write the contents of the xmlfile to 'content.xml'
  rc = write_rgn_to_file((const char*)file->buf->content, SIZE_MAX, 
                         file->tmpdir, "content.xml");
  if (rc < 0)
    return rc;

  // tar up the contents
  const char* const argv[] = { "tar", "--create", "--gzip", 
			       "-f", outfile_name, 
			       "-C", file->tmpdir,
			       "content", "content.xml", 0 };
  rc = runv("/bin/tar", argv);
  if (rc < 0)
    return rc;

  return 0;
}


//
// Return the contents of the reportfile as a string.
//
// Currently only works if 'file' is a signature.
//
char* 
reportfile_as_string( reportfile_t* file ) {

  if (0) {
    internal_error_printf("error: reportfile_as_string called before reportfile_end");
    return NULL;
  }

  if (!file->issignature) {
    internal_error_printf("error: reportfile_as_string called on non-signature");
    return NULL;
  }

  return strdup((const char*)file->buf->content);
}



//
// free the reportfile
//
int
reportfile_free( reportfile_t* file ) {
  int rc = 0;
    
  if (file) {
    // delete the temp dir
    if (file->tmpdir) {
      rc = remove_directory(file->tmpdir);
      if (rc < 0)
	rc = -1;
    }
    xmlFreeTextWriter(file->writer);
    xmlBufferFree(file->buf);
    free(file->tmpdir);
    free(file);
  }
  return rc;
}




static void
internal_error_curl(CURLcode err, const char* msg1, const char* msg2)
{
  internal_error_printf("error: %s: %s: %s", msg1, msg2, curl_easy_strerror(err));
}


typedef struct response_data {
  long code;
  const char* location;
  const char* strata_message;
  const char* http_header;
  const char* content_type;
  char* body;
  size_t body_size;
} response_data_t;

static void 
response_data_free(const response_data_t* response_data) {
  if (response_data) {
    free((void*)response_data->http_header);
    free((void*)response_data->strata_message);
    free((void*)response_data->content_type);
    free((void*)response_data->body);
    free((void*)response_data->location);
    free((void *)response_data);
  }
}  


// 
// Examine each header looking for "Location:" or "Strata-Message:" headers
// 
static size_t 
headerfunction(void *buffer_pv, size_t count, size_t nmemb, void *response_data_pv)
{
  size_t size = count * nmemb;
  struct response_data* response_data = (struct response_data*)response_data_pv;
  const char* buffer = (const char*)buffer_pv;

  const char location_key[] = "Location:";
  const size_t location_key_size = sizeof(location_key)-1;

  if (size >= location_key_size
      && 0 == memcmp(buffer,location_key,location_key_size)) {
    const char* start = (const char*) buffer+location_key_size+1;
    const char* end;

    // skip over any leading space
    while (start < buffer+size 
           && isspace(*start)) ++start;

    end = start;

    // skip till we find the end of the url (first space or end of buffer)
    while (end < buffer+size 
           && !isspace(*end)) ++end;

    response_data->location = strndup(start,end-start);
  }


  const char strata_message_key[] = "Strata-Message:";
  const size_t strata_message_key_size = sizeof(strata_message_key)-1;

  // We are going to stop using strata_message, 
  //   but for now only do minimal changes to stop using them
  if (0 
      && size >= strata_message_key_size
      && 0 == memcmp(buffer,strata_message_key,strata_message_key_size)) {
    const char* tmp_msg = (const char*) buffer+strata_message_key_size;
    const char* tmp_end = (const char*) buffer+size;

    while (tmp_msg < tmp_end && isspace(*tmp_msg)) 
      tmp_msg++;

    while (tmp_msg < tmp_end && isspace(*(tmp_end-1)))
      tmp_end--;

    size_t tmp_msg_len = tmp_end - tmp_msg;

    if (response_data->strata_message) {
      const char* old_msg = response_data->strata_message;
      const size_t old_msg_len = strlen(old_msg);
      const size_t new_msg_len = old_msg_len + tmp_msg_len + 1;
      char* new_msg = (char*)malloc(new_msg_len+1);

      mempcpy(mempcpy(mempcpy(new_msg, old_msg, old_msg_len), 
                              " ", 1), 
                              tmp_msg, tmp_msg_len);
      new_msg[new_msg_len] = 0;
      response_data->strata_message = new_msg;
      free((void*)old_msg);
    } else {
      response_data->strata_message = strndup(tmp_msg, tmp_msg_len);
    }
  }

  const char content_type_key[] = "Content-Type:";
  const size_t content_type_key_size = sizeof(content_type_key)-1;

  if (size >= content_type_key_size
      && 0 == memcmp(buffer,content_type_key,content_type_key_size)) {
    const char* tmp_msg = (const char*) buffer+content_type_key_size;
    const char* tmp_last = (const char*) buffer+size;

    while (tmp_msg < tmp_last && isspace(*tmp_msg)) 
      tmp_msg++;

    const char* tmp_end = tmp_msg;
    while (tmp_end < tmp_last
	   && !(isspace(*tmp_end) || ((*tmp_end) == ';')))
      tmp_end++;

    size_t tmp_msg_len = tmp_end - tmp_msg;

    if (tmp_msg_len) {
      if (response_data->content_type)
	acat_mem(&response_data->content_type, " ", 1);
      acat_mem(&response_data->content_type, tmp_msg, tmp_msg_len);
    }
  }

  const char http_header_key[] = "HTTP/";
  const size_t http_header_key_size = sizeof(http_header_key)-1;

  if (size >= http_header_key_size
      && 0 == memcmp(buffer,http_header_key,http_header_key_size)) {
    const char* tmp_msg = (const char*) buffer;
    const char* tmp_end = (const char*) buffer+size;

    while (tmp_msg < tmp_end && isspace(*(tmp_end-1)))
      tmp_end--;

    size_t tmp_msg_len = tmp_end - tmp_msg;
    response_data->http_header = strndup(tmp_msg, tmp_msg_len);
  }

  return size;
}

static int
append_accept_language_header(struct curl_slist ** httpheader_list_ptr) {
  int retval = -1;
  char* header = NULL;
  char* message_locale = NULL;

  // get the current MESSAGES locale
  char* environ_locale = setlocale(LC_MESSAGES, NULL);


  // catch all the locales we don't care about
  if (environ_locale == NULL 
      || 0 == strcmp("", environ_locale)
      || 0 == strcmp("C", environ_locale)
      || 0 == strcmp("POSIX", environ_locale))
    return 0;


  // Translate the environ locale to an http locale
  //   strip off everything after the initial list of alpha-numeric-underscores
  //   change underscores to dashes
  message_locale = strdup(environ_locale);
  if (!message_locale) {
    internal_error_printf("strdup failed");
    goto free_and_return;
  }
  char* p = NULL;
  for (p = message_locale; *p != 0 && (*p == '_' || isalnum(*p)); p++) {
    if (*p == '_')
      *p = '-';
  }
  *p = 0;

  // append the header to the header list
  header = ssprintf("Accept-Language: %s", message_locale);
  if (!header)
    goto free_and_return;
  *httpheader_list_ptr = curl_slist_append(*httpheader_list_ptr, header);
  retval = 0;

 free_and_return:;
  free(header);
  free(message_locale);
  return retval;
}
  
static int
append_accept_header(struct curl_slist ** httpheader_list_ptr) {
  // append the header to the header list
  char* header = strdup("Accept: text/plain");
  if (!header)
    return -1;
  *httpheader_list_ptr = curl_slist_append(*httpheader_list_ptr, header);

  free(header);
  return 0;
}

  
    
//
// URL: to post to
//
// content_type: of the data
// data: depends on data_size
//   if data_size is 0 or more
//      data is binary data to be sent in post
//   if data_size == -1 
//      data is NULL terminated string to be sent in post
//   if data_size == -2 
//      data is name of file whose contents are to be sent in post
//   if data_size == -3 
//      data is name of file whose contents are to be sent in post
//      the file is sent as a multipart/form-data post
//


static response_data_t* 
post(const char* URL,
     const char* cert_data,
     const char* username,
     const char* password,
     const char* content_type,
     const char* data,
     const long data_size) {

  CURLcode rc;
  CURL *handle = curl_easy_init();
  struct curl_slist *httpheader_list = NULL;
  struct curl_httppost* post = NULL; 
  FILE* data_file = NULL;
  FILE* body_stream = NULL;

  response_data_t* response_data = NULL;

  rc = curl_easy_setopt(handle, CURLOPT_VERBOSE, 0);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_VERBOSE)");
    goto free_and_return;
  }
  rc = curl_easy_setopt(handle, CURLOPT_NOPROGRESS, 1);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_NOPROGRESS)");
    goto free_and_return;
  }


  rc = curl_easy_setopt(handle, CURLOPT_POST, 1);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POST)");
    goto free_and_return;
  }

  rc = curl_easy_setopt(handle, CURLOPT_URL, URL);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_URL)");
    goto free_and_return;
  }

  if (username) {
    rc = curl_easy_setopt(handle, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HTTPAUTH, CURLAUTH_BASIC)");
      goto free_and_return;
    }

#if defined(CURLOPT_USERNAME) && defined(CURLOPT_PASSWORD)
    rc = curl_easy_setopt(handle, CURLOPT_USERNAME, username);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_USERNAME)");
      goto free_and_return;
    }
    
    rc = curl_easy_setopt(handle, CURLOPT_PASSWORD, (password ? password : ""));
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_PASSWORD)");
      goto free_and_return;
    }
#else
    {
      const char* userpwd = ssprintf("%s:%s", username, 
                                     (password ? password : ""));
      if (!userpwd) goto free_and_return;
      rc = curl_easy_setopt(handle, CURLOPT_USERPWD, userpwd);
      free((void*)userpwd);
      if (rc) {
        internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_USERPWD)");
        goto free_and_return;
      }
    }
#endif
  }

  if (cert_data) {
    if (!strcmp(cert_data,"INSECURE")) {
      rc = curl_easy_setopt(handle, CURLOPT_SSL_VERIFYPEER, 0);
      if (rc) {
	internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_SSL_VERIFYPEER)");
	goto free_and_return;
      }
      rc = curl_easy_setopt(handle, CURLOPT_SSL_VERIFYHOST, 0);
      if (rc) {
	internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_SSL_VERIFYHOST)");
	goto free_and_return;
      }
    }
  }

  data_file = 0;
  post = NULL;  
  struct curl_httppost* last = NULL;  

  if (data_size == -2) {
    data_file = fopen(data, "r");
    if (!data_file) {
      internal_error_printf("Can't open file '%s'", data);
      goto free_and_return;
    }
    rc = curl_easy_setopt(handle, CURLOPT_READDATA, data_file);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_READDATA)");
      goto free_and_return;
    }

  } else if (data_size == -3) {
    rc = curl_formadd(&post, &last, 
                      CURLFORM_PTRNAME, "file", 
                      CURLFORM_FILE, data, 
                      CURLFORM_CONTENTTYPE, content_type,
                      CURLFORM_FILENAME, data,
                      CURLFORM_END);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_formadd(CURLFORM_FILE)");
      goto free_and_return;
    }
    
    rc = curl_easy_setopt(handle, CURLOPT_HTTPPOST, post);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HTTPPOST);");
      goto free_and_return;
    }

  } else {
    rc = curl_easy_setopt(handle, CURLOPT_POSTFIELDS, data);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POSTFIELDS)");
      goto free_and_return;
    }
    rc = curl_easy_setopt(handle, CURLOPT_POSTFIELDSIZE, data_size);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POSTFIELDSIZE)");
      goto free_and_return;
    }
  }


  httpheader_list = NULL;

  if (data_size != -3) {
    char* content_type_header = ssprintf("Content-Type: %s", content_type);
    if (!content_type_header) goto free_and_return;
    httpheader_list = curl_slist_append(httpheader_list,content_type_header);
    free(content_type_header);
  }

  rc = append_accept_language_header(&httpheader_list);
  if (rc) goto free_and_return;

  rc = append_accept_header(&httpheader_list);
  if (rc) goto free_and_return;

  curl_easy_setopt(handle, CURLOPT_HTTPHEADER, httpheader_list);


  rc = curl_easy_setopt(handle, CURLOPT_HEADERFUNCTION, headerfunction);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HEADERFUNCTION)");
    goto free_and_return;
  }

  response_data = (response_data_t*)calloc(1,sizeof(response_data_t));
  if (!response_data) {
    internal_error_printf("strata_client: calloc return NULL\n");
    goto free_and_return;
  }

  rc = curl_easy_setopt(handle, CURLOPT_WRITEHEADER, response_data );
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_WRITEHEADER)");
    free((void*)response_data);
    response_data = NULL;
    goto free_and_return;
  }

  body_stream = open_memstream(&response_data->body, &response_data->body_size); 
  rc = curl_easy_setopt(handle, CURLOPT_WRITEDATA, body_stream);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_WRITEDATA)");
    free((void*)response_data);
    response_data = NULL;
    goto free_and_return;
  }

  rc = curl_easy_perform(handle);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_perform");
    free((void*)response_data);
    response_data = NULL;
    goto free_and_return;
  }

  rc = curl_easy_getinfo(handle, CURLINFO_RESPONSE_CODE, &response_data->code); 
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_getinfo(CURLINFO_RESPONSE_CODE)");
    free((void*)response_data);
    response_data = NULL;
    goto free_and_return;
  }

 free_and_return:;
  if (body_stream) fclose(body_stream);
  if (data_file) fclose(data_file);
  if (post) curl_formfree(post);
  if (httpheader_list) curl_slist_free_all(httpheader_list);
  curl_easy_cleanup(handle);

  return response_data;
}

static response_data_t* 
post_string(const char* URL,
	    const char* cert_data,
            const char* username,
            const char* password,
            const char* content_type,
            const char* str) {
  return post(URL, cert_data, username, password, content_type, str, -1);
}

static response_data_t* 
post_memory(const char* URL,
	    const char* cert_data,
            const char* username,
            const char* password,
            const char* content_type,
            const char* data,
            const long data_size)
  __attribute__ ((unused));
static response_data_t* 
post_memory(const char* URL,
	    const char* cert_data,
            const char* username,
            const char* password,
            const char* content_type,
            const char* data,
            const long data_size) {
  if (data_size < 0) {
    internal_error_printf("data size is negative");
    return NULL;
  }

  return post(URL, cert_data, 
	      username, password, content_type, data, data_size);
}

static response_data_t* 
post_namedfile(const char* URL,
	       const char* cert_data,
               const char* username,
               const char* password,
               const char* content_type,
               const char* namedfile)
  __attribute__ ((unused));
static response_data_t* 
post_namedfile(const char* URL,
	       const char* cert_data,
               const char* username,
               const char* password,
               const char* content_type,
               const char* namedfile) {
  return post(URL, cert_data, username, password, content_type, namedfile, -2);
}

static response_data_t* 
postform_namedfile(const char* URL,
		   const char* cert_data,
                   const char* username,
                   const char* password,
                   const char* content_type,
                   const char* namedfile) {
  return post(URL, cert_data, username, password, content_type, namedfile, -3);
}











static const char* 
make_case_data(const char* summary, const char* description,
               const char* product, const char* version,
               const char* component) {
  const char* retval = NULL;
  xmlTextWriterPtr writer = NULL;
  xmlBufferPtr buf = NULL;
  int rc;

  buf = xmlBufferCreate();
  if (buf == NULL) {
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    goto free_and_return;
  }
  writer = xmlNewTextWriterMemory(buf, 0);
  if (writer == NULL) {
    internal_error_printf("strata_client: Error creating the xml writer\n");
    goto free_and_return;
  }

  // start a new xml document
  rc = xmlTextWriterStartDocument(writer, NULL, "UTF-8", "yes");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    goto free_and_return;
  }

  // start the document element 'case'
  rc = xmlTextWriterStartElement(writer, (const xmlChar*)"case");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    goto free_and_return;
  }

  // add the 'xmlns' attribue
  rc = xmlTextWriterWriteAttribute(writer, (const xmlChar*)"xmlns",
                                   (const xmlChar*)"http://www.redhat.com/gss/strata");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    goto free_and_return;
  }

  // open a binding element
  rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"summary", (const xmlChar*)summary);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteElement\n");
    goto free_and_return;
  }

  rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"description", (const xmlChar*)description);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteElement\n");
    goto free_and_return;
  }

  if (product) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"product", (const xmlChar*)product);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      goto free_and_return;
    }
  }

  if (version) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"version", (const xmlChar*)version);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      goto free_and_return;
    }
  }

  if (component) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"component", (const xmlChar*)component);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      goto free_and_return;
    }
  }

  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(writer);
  if (rc < 0) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    goto free_and_return;
  }

  retval = strdup((const char*)buf->content);

 free_and_return:;
  xmlFreeTextWriter(writer);
  xmlBufferFree(buf);
  
  return retval;
}

static const char* 
make_response_xml(const char* title, const char* body, 
                  const char* actualURL, const char* displayURL) {
  const char* retval = NULL;
  xmlTextWriterPtr writer = NULL;
  xmlBufferPtr buf = NULL;
  int rc;
 
  buf = xmlBufferCreate();
  if (buf == NULL) {
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    goto free_and_return;
  }
  writer = xmlNewTextWriterMemory(buf, 0);
  if (writer == NULL) {
    internal_error_printf("strata_client: Error creating the xml writer\n");
    goto free_and_return;
  }
  
  // start a new xml document
  rc = xmlTextWriterStartDocument(writer, NULL, "UTF-8", "yes");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    goto free_and_return;
  }

  rc = xmlTextWriterStartElement(writer, (const xmlChar*)"response");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    goto free_and_return;
  }

  if (title) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"title", (const xmlChar*)title);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      goto free_and_return;
    }
  }

  if (body) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"body", (const xmlChar*)body);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      goto free_and_return;
    }
  }
    
  if (actualURL || displayURL) {
    rc = xmlTextWriterStartElement(writer, (const xmlChar*)"URL");
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterStartElement\n");
      goto free_and_return;
    }

    if (actualURL) {
      rc = xmlTextWriterWriteAttribute(writer, (const xmlChar*)"href", (const xmlChar*)actualURL);
      if (rc < 0) {
        internal_error_printf
          ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        goto free_and_return;
      }
    }

    if (displayURL) {
      rc = xmlTextWriterWriteString(writer, (const xmlChar*)displayURL);
      if (rc < 0) {
        internal_error_printf
          ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        goto free_and_return;
      }
    }
  }
  
  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(writer);
  if (rc < 0) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    goto free_and_return;
  }

  retval = strdup((const char*)buf->content);


 free_and_return:;
  xmlFreeTextWriter(writer);
  xmlBufferFree(buf);
  
  return retval;
}

static int
acat_xml_as_form(const char** oldp, int indent, xmlNode * a_node) {
  xmlNode* cur_node;
  int x;

  if (a_node->type == XML_ELEMENT_NODE) {
    if (a_node->name) {
      if (acat_str(oldp, "\n")) return 1;
      for (x = 0; x < indent; x++) if (acat_str(oldp, " ")) return 1;
      if (acat_str(oldp, (const char*)a_node->name)) return 1;
      if (acat_str(oldp, ": ")) return 1;
      indent += 2;
    }
    for (cur_node = a_node->children; cur_node; cur_node = cur_node->next)
      if (acat_xml_as_form(oldp, indent, cur_node)) return 1;
  }
  else if (a_node->type == XML_TEXT_NODE
	   || a_node->type == XML_CDATA_SECTION_NODE) { 
    if (a_node->content) {
      int len = strlen((const char*)a_node->content);
      
      const char* start = (const char*)a_node->content;
      const char* end = (const char*)(a_node->content + len);
      while (start < end && isspace(*start)) start++;
      while (start < (end-1) && isspace(*(end-1))) end--;
      if (acat_mem(oldp, start, end-start))
	return 1;
    }
    for (cur_node = a_node->children; cur_node; cur_node = cur_node->next)
      if (acat_xml_as_form(oldp, indent, cur_node)) return 1;
  }
  else {
    for (cur_node = a_node->children; cur_node; cur_node = cur_node->next)
      if (acat_xml_as_form(oldp, indent, cur_node)) return 1;
  }

  return 0;
}


static int
xmlChar_is(const xmlChar* a, const char* b) {
  return !strcmp((const char*)a, (const char*)b);
}

static int
acat_html_list_as_text(const char** oldp, xmlNode* list) {
  xmlNode *cur_node = NULL;

  for (cur_node = list; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if (xmlChar_is(cur_node->name,"p")
	  || xmlChar_is(cur_node->name,"div")) {
	if (acat_str(oldp, "\n")) return 1;
	if (cur_node->children) 
	  if (acat_html_list_as_text(oldp, cur_node->children))
	    return 1;
      } 
      else if (xmlChar_is(cur_node->name,"h1")
	       || xmlChar_is(cur_node->name,"h2")
	       || xmlChar_is(cur_node->name,"h3")
	       || xmlChar_is(cur_node->name,"h4")
	       || xmlChar_is(cur_node->name,"h5")) {
	if (acat_str(oldp, "\n")) return 1;
	if (cur_node->children) {
	  if (acat_html_list_as_text(oldp, cur_node->children))
	    return 1;
	  if (acat_str(oldp, "\n")) return 1;
	}
      }
      else {	
	if (cur_node->children) 
	  if (acat_html_list_as_text(oldp, cur_node->children))
	    return 1;
      }
    }
    else if (cur_node->type == XML_TEXT_NODE
	     || cur_node->type == XML_CDATA_SECTION_NODE) {
      if (acat_str(oldp, (const char*)cur_node->content)) return 1;
      if (cur_node->children) 
	if (acat_html_list_as_text(oldp, cur_node->children))
	  return 1;
    }
    else {
      if (cur_node->children) 
	if (acat_html_list_as_text(oldp, cur_node->children))
	  return 1;
    }
  }
  return 0;
}

static const char*
translate_body(response_data_t* response_data) {

  if (response_data->content_type
      && strlen(response_data->content_type)) {
    if (!strcmp(response_data->content_type,"text/plain")) {
      return strdup(response_data->body);
    }
    else if (!strcmp(response_data->content_type,"text/xml")) {
      const char *p;
      const char *p_end = response_data->body + strlen(response_data->body);
      for (p = response_data->body; 
	   p < p_end && isspace(*p); 
	   p++);
      if (*p != '<') {
	return ssprintf("Content of type %s did not start with '<':\n%s",
			response_data->content_type,
			response_data->body);
      }

      xmlDocPtr doc = xmlParseMemory(response_data->body,
				     strlen(response_data->body));
      if (!doc) {
	xmlCleanupParser();
	return ssprintf("Could not parse content of type %s as XML: \n"
		       "%s", 
		       response_data->content_type,
		       response_data->body);
      }

      const char* fmt = NULL;
      if (acat_xml_as_form(&fmt, 0, xmlDocGetRootElement(doc))) {
	xmlFreeDoc(doc);
	xmlCleanupParser();
	free((void*)fmt);
	return NULL;
      }
      xmlFreeDoc(doc);
      xmlCleanupParser();
      if (!fmt)
	return ssprintf("Could not translate content type of '%s' to text:\n%s",
			response_data->content_type, response_data->body);
      const char* rtn = ssprintf(
			 "Translated '%s' response to text:\n%s\n"
		         "Original response of content type %s:\n%s",
			 response_data->content_type, fmt, 
			 response_data->content_type, response_data->body);
      free((void*)fmt);
      return rtn;
    }
    else if (!strcmp(response_data->content_type,"text/html")) {
      const char *p;
      const char *p_end = response_data->body + strlen(response_data->body);
      for (p = response_data->body; 
	   p < p_end && isspace(*p); 
	   p++);
      if (*p != '<') {
	return ssprintf("Content of type '%s' did not start with '<':\n%s",
			response_data->content_type,
			response_data->body);
      }

      xmlDocPtr doc = htmlReadMemory(response_data->body,
				     strlen(response_data->body),
				     NULL, NULL, 0);
      if (!doc) {
	xmlCleanupParser();
	return ssprintf("Could not parse content of type '%s' as HTML: \n"
			"%s", 
			response_data->content_type,
			response_data->body);
      }
      const char* fmt = NULL;
      xmlNode *cur_node = NULL;

      for (cur_node = xmlDocGetRootElement(doc)->children; 
	   cur_node; 
	   cur_node = cur_node->next) {
	if (cur_node->type == XML_ELEMENT_NODE 
	    && xmlChar_is(cur_node->name,"body")) {
	  if (acat_html_list_as_text(&fmt, cur_node->children)) {
	    xmlFreeDoc(doc);
	    xmlCleanupParser();
	    free((void*)fmt);
	    return NULL;
	  }
	}
      }
      xmlFreeDoc(doc);
      xmlCleanupParser();
      if (!fmt) 
	return ssprintf("Could not translate content type of '%s' to text:\n%s",
			response_data->content_type, response_data->body);
      const char* rtn = ssprintf(
                          "Translated '%s' response to text:\n%s\n"
			  "Original response of content type %s:\n%s",
			  response_data->content_type, fmt, 
			  response_data->content_type, response_data->body);
      free((void*)fmt);
      return rtn;
    }
    else {
      const char* rtn;
      if (!strcmp(response_data->content_type,"text")
	  || !strncmp(response_data->content_type,"text/",strlen("text/"))) {
	rtn = ssprintf("Server returned unexpected content type '%s':\n%s",
		       response_data->content_type, response_data->body);
      } 
      else {
	rtn = ssprintf("Server returned unexpected content type '%s'",
		       response_data->content_type);
      }

      return rtn;
    }
  }
  else {
    const char* rtn = 
      ssprintf("Server returned content with no content type header:\n%s",
	       response_data->body);
    return rtn;
  }
}


static const char*
append_http_response_message(const char* body, 
                             response_data_t* response_data) {

  const char* status = response_data->http_header;
  const char* generated_status = NULL;
  if (!status) {
    generated_status = ssprintf("HTTP Response code %ld", response_data->code);
    if (!generated_status) return NULL;
    status = generated_status;
  }

  if (response_data->code == 200 || response_data->code == 201) {
    /* valid and expected codes, don't add any message to 'body' */
    free((void*)generated_status);
  }

  else if (200 <= response_data->code && response_data->code < 300) {
    const char* new_body = ssprintf("%s%s"
        "%s\n"
        "Server returned a successful response code other than "
            "those expect by this client",
                                    body,
                                    (*body ? "\n" : ""), 
                                    status);
    free((void*)generated_status);
    if (!new_body) return NULL;
    if (new_body != body) {
      free((char*)body);
      body = new_body;
    }
  }
  
  else if (300 <= response_data->code && response_data->code < 400) {
    const char* location = response_data->location;
    if (!location || strlen(location) == 0) {
      const char* new_body = ssprintf(
               "%s%s"
	       "Redirect without Location Header\n"
	       "%s\n"
	       "Server returned a redirect response code, but the "
	       "response included no Location header to redirect to",
	       body,
	       (*body ? "\n" : ""), 
	       status);
      free((void*)generated_status);
      if (!new_body) return NULL;
      if (new_body != body) {
	free((char*)body);
	body = new_body;
      }
    } else if (response_data->code == 305) {
      // we handle 305 up to some fixed number (currently 10) of
      // redirects, after that we error
      const char* new_body = ssprintf(
             "%s%s"
             "Too Many Redirects\n"
	     "%s\n"
	     "This client has been redirected more times than it can handle\n"
	     "   The latest redirect was to:\n"
	     "      %s\n",
	     body,
	     (*body ? "\n" : ""), 
	     status,
	     location);
      free((void*)generated_status);
      if (!new_body) return NULL;
      if (new_body != body) {
	free((char*)body);
	body = new_body;
      }
    } else {
      const char* new_body = ssprintf(
		  "%s%s"
		  "Unhandled Redirect\n"
		  "%s\n"
		  "Server returned a redirect response code that this "
		  "client does not automatically handle\n"
		  "The server is redirecting this request to:\n"
		  "    %s\n"
		  "If you wish you may change your strata plugin configuration "
		  "to point to this URL",
		  body,
		  (*body ? "\n" : ""), 
		  status,
		  location);
      free((void*)generated_status);
      if (!new_body) return NULL;
      if (new_body != body) {
	free((char*)body);
	body = new_body;
      }
    }
  }
  
  else if (400 <= response_data->code && response_data->code < 500) {
    const char* new_body = ssprintf("%s%s"
                                    "%s",
                                    body,
                                    (*body ? "\n" : ""), 
                                    status);
    free((void*)generated_status);
    if (!new_body) return NULL;
    if (new_body != body) {
      free((char*)body);
      body = new_body;
    }
  }
  
  else if (500 <= response_data->code && response_data->code < 600) {
    const char* new_body = ssprintf("%s%s"
                                    "Server Internal Error\n"
                                    "%s",
                                    body,
                                    (*body ? "\n" : ""), 
                                    status);
    free((void*)generated_status);
    if (!new_body) return NULL;
    if (new_body != body) {
      free((char*)body);
      body = new_body;
    }
  }
  
  else {
    const char* new_body = ssprintf("%s%s"
            "Unexpected Response Code\n"
            "%s\n"
            "Server returned a response code that this client does not handle",
                                    body,
                                    (*body ? "\n" : ""), 
                                    status);
    free((void*)generated_status);
    if (!new_body) return NULL;
    if (new_body != body) {
      free((char*)body);
      body = new_body;
    }
  }

  return body;
}

static const char*
append_response_header(const char* body, const char* action) {
  const char* new_body = ssprintf("%s%sResponse for %s:",
                                  body,
                                  (*body ? "\n" : ""), 
                                  action);
  if (!new_body) return NULL;
  if (new_body != body) {
    free((char*)body);
    body = new_body;
  }
  return body;
}


static const char*
append_response_message(const char* body, 
                        const int header_needed,
                        const char* action, 
                        response_data_t* response_data) {

  // This function and append_http_response_message need to be
  // refactored.  Leaving them as they are for now to simplify the
  // patches that are going to be applied to the RHEL6 release.

  // a header is only needed if we are combining responses for two POSTS
  int header_added = ! header_needed;  
  
  // For 200 or 201 we only print the body (the content of the response)
  // For 400 <= x < 500 we print the header after the body
  if (! (response_data->code == 200 
         || response_data->code == 201
         || (400 <= response_data->code && response_data->code < 500))) {
    if (!header_added) {
      body = append_response_header(body, action);
      if (!body) return NULL;
      header_added = 1;
    }
    body = append_http_response_message(body, response_data);
    if (!body) return NULL;
    
    // We are going to stop using strata_message, 
    //   but for now only do minimal changes to stop using them
    if (0 && response_data->strata_message) {
      const char* new_body = ssprintf("%s%sStrata Server Message: %s",
                                      body,
                                      (*body ? "\n" : ""), 
                                      response_data->strata_message);
      if (!new_body) return NULL;
      if (new_body != body) {
        free((char*)body);
        body = new_body;
      }
    }
  }
  
  if (response_data->body 
      && strlen(response_data->body)) {
    if (!header_added) {
      body = append_response_header(body, action);
      if (!body) return NULL;
      header_added = 1;
    }

    const char* translated_body = translate_body(response_data);
    if (!translated_body) return NULL;

    const char* new_body = ssprintf("%s%s%s",
                                    body,
                                    (*body ? "\n" : ""), 
                                    translated_body);
    free((char*)translated_body);
    if (!new_body) return NULL;
    if (new_body != body) {
      free((char*)body);
      body = new_body;
    }
  }

  return body;
}
  
static const char* 
make_response(const char* action1,
              const char* action2,
              response_data_t* first_response,
              response_data_t* second_response,
              const char* display_url) {

  const char* retval = NULL;
  int success = 0;

  // It has been suggested that we should not output internal
  // headers, even for cases where there are two requests
  // Leave the code till we are sure one way or another.
  int header_needed = 0;  // (second_response != NULL);
 
  if (200 <= first_response->code
      && first_response->code < 300)
    success = 1;

  const char* title;
  if (success)
    title = ssprintf("%s Succeeded", action1);
  else 
    title = ssprintf("%s Failed", action1);
  if (!title) return NULL;

  const char* body = strdup("");
  if (!body) goto free_and_return;

  body = append_response_message(body, header_needed,
				 action1, first_response);
  if (!body) goto free_and_return;
  
  if (second_response) {

    // It has been suggested that we should only talk about 
    // the first response title.
    // Leave the code till we are sure one way or another.
    if (0) {
      const char* new_title;

      if (200 <= second_response->code
	  && second_response->code < 300)
	new_title = ssprintf("%s; %s Succeeded", title, action2);
      else 
	new_title = ssprintf("%s; %s Failed", title, action2);

      if (!new_title) goto free_and_return;
      if (new_title != title) {
	free((char*)title);
	title = new_title;
      }
    }

    body = append_response_message(body, header_needed,
				   action2, second_response);
    if (!body) goto free_and_return;
  }

  // if we have been successful, append the 'name' of the POST-ed
  // thing to the passed in display_url; where we define 'name' as
  // the string after the last '/' in the actual URL

  const char* displayURL = NULL;
  if (success 
      && first_response->location 
      && strlen( first_response->location )) {
    const char* start = first_response->location;
    const char* end = start + strlen(start);
    while (start < end && *(end-1) == '/')
      end--;
    const char* p = end;
    while (start < p && *(p-1) != '/')
      p--;
    if (start < p && p < end) {
      if (acat_str(&displayURL, display_url)) goto free_and_return;
      if (acat_str(&displayURL, " ")) goto free_and_return;
      if (acat_mem(&displayURL, p, end - p)) goto free_and_return;
      display_url = displayURL;
    }
  }

  retval = make_response_xml(title, body,
			     (success ? first_response->location : NULL),
			     (success ? display_url : NULL));

 free_and_return:;
  free((char*)displayURL);
  free((char*)body);
  free((char*)title);
  return retval;
}

static response_data_t* 
attach_file_to_case( const char* baseURL,
		     const char* cert_data,
                     const char* username,
                     const char* password,
                     const char* case_name,
                     const char* report_file_name ) {

  const char* attachURL;

  // if case_name is NULL, treat baseURL as caseURL
  if (case_name) {
    const char* caseURL = concat_path_file(baseURL, "/cases");
    if (!caseURL) return NULL;

    const char* freethis = caseURL;
    caseURL = concat_path_file(caseURL, case_name);
    free((void*)freethis);
    if (!caseURL) return NULL;

    attachURL = concat_path_file(caseURL, "/attachments");
    free((void*)caseURL);
    if (!attachURL) return NULL;
  } else {
    attachURL = concat_path_file(baseURL, "/attachments");
    if (!attachURL) return NULL;
  }
  
  int redirect_attach_count = 0;
 redirect_attach:;
  response_data_t* attach_response = 
    postform_namedfile(attachURL, cert_data,
                       username, password,
                       "application/binary", 
                       report_file_name);
  if (!attach_response) goto free_and_return;
  
  if (attach_response->code == 305
      && attach_response->location 
      && (++redirect_attach_count <= 10)) {
    free((void*)attachURL);
    attachURL = attach_response->location;
    attach_response->location = NULL;
    response_data_free(attach_response);
    goto redirect_attach;
  }
    
 free_and_return:;
  free((void*)attachURL);
  return attach_response;
}

const char* 
post_signature(const char* baseURL,
	       const char* cert_data,
               const char* username,
               const char* password,
               const char* signature) {
  const char* URL = concat_path_file(baseURL, "/signatures");
  if (!URL) return NULL;

  int redirect_signature_count = 0;
 redirect_signature:;
  response_data_t* response_data = post_string(URL, cert_data,
                                               username,
                                               password,
                                               "application/xml",
                                               signature);
  if (!response_data) goto free_and_return;

  const char* retval = NULL;
  switch (response_data->code) {
  case 200:
  case 201:
    if (response_data->body && strlen(response_data->body) > 0) {
      retval = response_data->body;
      response_data->body = NULL;
    }
    else if (response_data->strata_message 
             && strcmp(response_data->strata_message,"CREATED") != 0) {
      retval = response_data->strata_message;
      response_data->strata_message = NULL;
    }
    else {
      retval = strdup("Thank you for submitting your signature.");
    }
    break;

  case 305:
    if (response_data->location 
	&& (++redirect_signature_count <= 10)) {
      free((void*)URL);
      URL = response_data->location;
      response_data->location = NULL;
      response_data_free(response_data);
      goto redirect_signature;
    }

    retval = make_response("Signature Submission", "",
			     response_data, NULL,
			     "New Signature");
    break;

  default:
    retval = make_response("Signature Submission", "",
                           response_data, NULL,
                           "New Signature");
    break;
  }

 free_and_return:;
  response_data_free(response_data);
  free((void *)URL);
  return retval;
}

response_data_t* 
create_case(const char* baseURL,
	    const char* cert_data,
            const char* username,
            const char* password,
            const char* case_data_type,
            const char* case_data) {

  const char* URL = concat_path_file(baseURL, "/cases");
  if (!URL) return NULL;

  int redirect_createcase_count = 0;
 redirect_createcase:;
  response_data_t* createcase_response = post_string(URL, cert_data,
                                                     username,
                                                     password,
                                                     case_data_type,
                                                     case_data);
  if (!createcase_response) goto free_and_return;

  if (createcase_response->code == 305
      && createcase_response->location
      && (++redirect_createcase_count <= 10)) {
    free((void*)URL);
    URL = createcase_response->location;
    createcase_response->location = NULL;
    response_data_free(createcase_response);
    goto redirect_createcase;
  }
  
 free_and_return:;
  free((void*)URL);
  return createcase_response;
}


const char* 
send_report_to_new_case( const char* baseURL,
                         const char* cert_data,
			 const char* username,
                         const char* password,
                         const char* summary,
                         const char* description,
                         const char* component,
                         const char* product,
                         const char* version,
                         const char* report_file_name ) {

  const char* case_data = make_case_data(summary, description,
                                         product, version,
                                         component);
  if (!case_data) return NULL;

  response_data_t* createcase_response = create_case(baseURL,
						     cert_data,
                                                     username,
                                                     password,
                                                     "application/xml",
                                                     case_data);
  free((void*)case_data);
  if (!createcase_response) return NULL;

  const char* retval = NULL;
  if (200 <= createcase_response->code
      && createcase_response->code < 300) {

    if (createcase_response->location
        && strlen(createcase_response->location) > 0) {
      response_data_t* attach_response = 
        attach_file_to_case(createcase_response->location, cert_data, 
                            username, password,
                            NULL,
                            report_file_name);
      if (attach_response)
	retval = make_response("Case Creation", "File Attachment",
			       createcase_response, attach_response,
			       "New Case");
      else 
	retval = NULL;
      response_data_free(attach_response);
    } else {
      /* Case Creation returned valid code, but no location */
      retval = ssprintf("Error: case creation return HTTP Code %ld, "
                        "but no Location URL header", 
                        createcase_response->code);
    }
  } else {
    retval = make_response("Case Creation", NULL,
                           createcase_response, NULL,
                           "New Case");
  }

  response_data_free(createcase_response);
  return retval;
}

const char* 
send_report_to_existing_case( const char* baseURL,
			      const char* cert_data,
                              const char* username,
                              const char* password,
                              const char* case_name,
                              const char* report_file_name ) {

  const char* retval = NULL;
  
  response_data_t* attach_response = 
    attach_file_to_case(baseURL, cert_data, 
                        username, password, 
                        case_name, report_file_name);
  if (!attach_response) goto free_and_return;
  

  retval = make_response("File Attachment", NULL,
			 attach_response, NULL,
			 "New Attachment");
  if (!retval) goto free_and_return;


 free_and_return:;
  response_data_free(attach_response);
  return retval;
}
